// io_pcst.cpp
//
// This vpat_io module reads in PC-Stitch files (version 2.0 and 4.0)
//
// PC-Stitch is available from http://www.pcstitch.com/, where some free 
// charts are periodically released.
//
// DMC (http://www.dmc-usa.com/) periodically releases free charts in this 
// file format as well.
//
// This version of the reader has been tested with many charts from the 
// above URLs.

//#define _MSP_DEBUG_

#include "vpat_io.h"
#include "io_pcst.h"

#ifdef MSDOS
#include <io.h>
#include <string.h>
#else
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#define O_BINARY 0

#endif

#include <stdio.h>
#include <fcntl.h>

#include <malloc.h>

#ifdef _MSP_DEBUG_
#include <v/vapp.h>
#include <v/vnotice.h>
#endif

//
// Is this a PCStitch file?
//
// true if yes, false if no
//
// Check is pretty simple: if the first byte of the header is 2 or 4, it's 
// an okay file.  However, since other file formats *also* start with 2 or 
// 4, loop over the description chunk, looking for NULLs.  Any NULLs so 
// encountered flag it as NOT a PC-Stitch file (since PC-Stitch descriptions 
// are of fixed length and are padded with spaces).
//
int IsPCStitch(char *chunk, int size)
{
    int i;

    if ((chunk[0]==2)||(chunk[0]==4))
      {
        for (i=0; i < 255; i++)
          if (chunk[i+9]==0) return false;              // no nulls allowed
        return true;
      }
    else
      {
        return false;
      }
}

// Some assumptions:
//     pat is checked to see if it's NULL
//     However, the internals are assumed to be NULL before calling, so 
//     the function calling this is responsible for freeing the pattern 
//     data before reuse
int ReadPCStitch(vpat_chart *pat, char *name)
  {

    int num_pal_entries;

    int num_good, num_bad;

    int pat_fd, num_read, num_extras, version,
        i, x, y, s;                                // looping variables
    unsigned int stitch_count;
    unsigned char in_data[38];

    if (!pat) return VPATIO_BADPAT;
    pat->palette = NULL;
    pat->stitches = NULL;
    pat->bstitches = NULL;

#ifdef _MSP_DEBUG_
    fprintf(stderr, "Starting to read PC-Stitch file...\n");
#endif

    if ((pat_fd = open(name, O_RDONLY|O_BINARY)) < 0)
      {
        return VPATIO_OPEN;
      }

    // no running off the end here! (this makes the strncpy safe)
    if (strlen(name) > 255) name[255]=0x00;
    strncpy(pat->header.fname, name, strlen(name));

#ifdef _MSP_DEBUG_
    fprintf(stderr, "copied name\n");
#endif

    num_read=read(pat_fd, in_data, 9);
    if (num_read < 9)
      {
        return VPATIO_SHORT;
      }

    version = in_data[0];

    switch (version)
      {
        case 2:
          num_pal_entries = 64;
          break;

        case 4:
          num_pal_entries = 100;
          break;

        default:
          return VPATIO_UNKNOWN;
      }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "version determined\n");
#endif

    pat->header.height        = (int)in_data[1] + (int)((int) in_data[2] << 8);
    pat->header.width         = (int)in_data[3] + (int)((int) in_data[4] << 8);
    pat->header.num_bstitches = (int)in_data[7] + (int)((int) in_data[8] << 8);

// now, we read the next 255 bytes in; this is the descriptive text which
// accompanies the pattern.

    sprintf((char *)&in_data[0], "Width: %d, Height: %d, size: %d", pat->header.width, pat->header.height,
            sizeof(vpat_stitch));

#ifdef _MSP_DEBUG_
    vNoticeDialog notice(theApp);
    (void)notice.Notice((char *)&in_data[0]);
#endif

    for (i=0 ; i<256; i++) pat->header.descr[i]=0;
    num_read=read(pat_fd, pat->header.descr, 255);
    if (num_read < 255)
      {
        return VPATIO_SHORT;
      }

    pat->header.num_colors=0;
    pat->palette = (vpat_color_entry *)malloc(sizeof(vpat_color_entry) *
                                              num_pal_entries);
    if (!pat->palette) return VPATIO_MALLOC;

#ifdef _MSP_DEBUG_
    fprintf(stderr, "Reading palette...");
#endif

    for (i=0 ; i < num_pal_entries ; i++)
      {
        num_read=read(pat_fd, in_data, 37);
        if (num_read < 37)
          {
            return VPATIO_SHORT;
          }
        pat->palette[i].used_flag = in_data[3];
        if (in_data[3] != 0xFF)                     // if it's used
          {
            pat->header.num_colors++;
            pat->palette[i].red = in_data[0];
            pat->palette[i].green = in_data[1];
            pat->palette[i].blue = in_data[2];
            pat->palette[i].DMC_num = (int)in_data[4] + 
                                      (int)((int)in_data[5] << 8);
            pat->palette[i].Anchor_num = (int)in_data[6] + 
                                         (int)((int)in_data[7] << 8);
            pat->palette[i].JPCoats_num = (int)in_data[8] + 
                                          (int)((int)in_data[9] << 8);
            strncpy(pat->palette[i].name, (char *)(&in_data[10]), 25);
            pat->palette[i].name[26] = '\0';       // just in case
            pat->palette[i].symbol = (int)in_data[35] + 
                                     (int)((int)in_data[36]);
            _colors[i].ResetColor(in_data[0],in_data[1],in_data[2]);
          }
      }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "read.\n");
#endif

    _bgcolor.ResetColor(255,255,255);

    pat->header.num_stitches = (unsigned int)pat->header.width * 
                               (unsigned int)pat->header.height;

    pat->stitches = (vpat_stitch *)malloc(sizeof(vpat_stitch) *
                                          pat->header.num_stitches);
    if (!pat->stitches)
      {
        return VPATIO_MALLOC;
      }

    num_read=2; stitch_count = 0;

#ifdef _MSP_DEBUG_
    fprintf(stderr, "reading stitches...");
#endif

    while ((num_read != 0) && 
           (stitch_count < pat->header.num_stitches)) {
      num_read = read(pat_fd, in_data, 2);
      if (num_read != 0) {
        for (i = 0; (i < in_data[0])&&(i < pat->header.num_stitches); i++)
        {
          if (in_data[1] == 255)   // pc-stitch magic number (no stitch here)
            {
              pat->stitches[stitch_count].color = -1;    // vpatview magic #
            }
          else
            {
              pat->stitches[stitch_count].color = in_data[1];
            }
          pat->stitches[stitch_count++].second = 0;
        }
      }
    }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "done.\nreading qstitches...");
#endif

    // okay, colors read in, now let's grab the quarter-stitch info
    num_read=2; stitch_count = 0;

    while ((num_read != 0) && 
           (stitch_count < pat->header.num_stitches)) {
      num_read = read(pat_fd, in_data, 2);
      if (num_read != 0) {
        for (i = 0; i < in_data[0]; i++)
        {
          pat->stitches[stitch_count++].type = in_data[1];
        }
      }
    }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "read.\n");
#endif

    // now, we read in the backstitch stuff (this is just as gross a hack 
    // as the rest of this)

#ifdef _MSP_DEBUG_
    fprintf(stderr, "reading bstitches...");
#endif

    if (pat->header.num_bstitches > 1)
      {

        pat->bstitches = (vpat_bstitch *)malloc((pat->header.num_bstitches) *
                                                (sizeof(vpat_bstitch)));

        if (!pat->bstitches)
          {
            return VPATIO_MALLOC;
          }

        switch (version)
          {
            case 2:                      // version 2
              for (i = 0 ; ((i < pat->header.num_bstitches)&&(num_read!=0)); 
                   i++)
                {

                  num_read = read(pat_fd, in_data, 8);

                  if (num_read == 8)
                    {
                      pat->bstitches[i].x1 = ((int)in_data[0] + 
                                              (int)((int) in_data[1] << 8))*2;
                      pat->bstitches[i].y1 = ((int)in_data[2] +
                                              (int)((int) in_data[3] << 8))*2;
                      pat->bstitches[i].x2 = ((int)in_data[4] +
                                              (int)((int) in_data[5] << 8))*2;
                      pat->bstitches[i].y2 = ((int)in_data[6] +
                                              (int)((int) in_data[7] << 8))*2;
                      pat->bstitches[i].color = -1;  // this format doesn't 
                                                     // understand color
                      pat->bstitches[i].type = 0;    // back stitch
                    }
    
                }
              break;

            case 4:                     // this is the new version
              for (i = 0 ; ((i < pat->header.num_bstitches)&&(num_read!=0));
                   i++)
                {

                  num_read = read(pat_fd, in_data, 10);

                  if (num_read == 10)
                    {
                      pat->bstitches[i].x1 = ((int)in_data[0] +
                                              (int)((int) in_data[1] << 8))*2;
                      pat->bstitches[i].y1 = ((int)in_data[2] +
                                              (int)((int) in_data[3] << 8))*2;
                      pat->bstitches[i].x2 = ((int)in_data[4] +
                                              (int)((int) in_data[5] << 8))*2;
                      pat->bstitches[i].y2 = ((int)in_data[6] +
                                              (int)((int) in_data[7] << 8))*2;
                      pat->bstitches[i].color=(int)in_data[8] +
                                             (int)((int) in_data[9] << 8);
                      pat->bstitches[i].type = 0;   // back stitch
                    }

                }

            default:
              break;

          }

      }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "read.\n");
#endif

    num_read = read(pat_fd, in_data, 2);
    num_extras = (int)in_data[0] + (int)((int) in_data[1] << 8);

#ifdef _MSP_DEBUG_
    fprintf(stderr, "extras(%d) ", num_extras);
#endif

    num_good=0; num_bad=0;
    for (i=0; i<num_extras; i++)
      {
        num_read = read(pat_fd, in_data, 6);
        if (num_read < 6)
          {
            close (pat_fd);
            return VPATIO_SHORT;
          }
        x = (int)in_data[0] + (int)((int) in_data[1] << 8);
        y = (int)in_data[2] + (int)((int) in_data[3] << 8);
        s = (int)in_data[4];
          // note that the 6th byte may be meaningful, but it doesn't have 
          // an obvious purpose, and it seems to be filled with garbage 
          // in some cases
        if ((x < pat->header.width) &&
            (y < pat->header.height) &&
            (s < pat->header.num_colors))
          {
            num_good++;
            pat->stitches[y*pat->header.width + x].second = s;
          }
        else
          {
            num_bad++;

#ifdef _MSP_DEBUG_
//          fprintf(stderr, "Unknown extra: [%d](%d,%d,%d)\n", i, x, y, s);
#endif

          }
      }

#ifdef _MSP_DEBUG_
    fprintf(stderr, "good: %d, bad: %d\n", num_good, num_bad);
#endif

    close(pat_fd);

#ifdef _MSP_DEBUG_
    fprintf(stderr, "read.\n");
#endif

    return VPATIO_OK;

  }

#ifdef _MIKE_

int WritePCStitch (vpat_chart *pat, char *name)
{
    int pat_fd, num_read, num_specials, color_bstitches, num_pal_entries, 
        i, j, x, y, testing;
    char out_data[256];

    if (!pat)
      {
        return VPATIO_BADPAT;
      }

    pat->header.num_stitches = pat->header.width * pat->header.height;

    if ((pat_fd=open(name, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY,0x01A4))<0)
      {
        return VPATIO_OPEN;
      }

    color_bstitches = true;
    for (i=0; i < pat->header.num_bstitches; i++)
      if (pat->bstitches[i].color == -1) color_bstitches = false;

    if (color_bstitches)
      {
        out_data[0] = 4;
        num_pal_entries = 100;
      }
    else
      {
        out_data[0] = 2;
        num_pal_entries = 64;
      }
    out_data[1] = (char)pat->header.height;
    out_data[2] = (char)(pat->header.height>>8);
    out_data[3] = (char)pat->header.width;
    out_data[4] = (char)(pat->header.width>>8);
    out_data[5] = 6;                     // magic value (unknown meaning)
    out_data[6] = 0;
    out_data[7] = (char)pat->header.num_bstitches;
    out_data[8] = (char)(pat->header.num_bstitches>>8);

    write(pat_fd, out_data, 9);

    for (i=0; i<256; i++) out_data[i]=0x20;    // fill with spaces
    write(pat_fd, out_data, 255);              // write mock description

    for (i=0; i < num_pal_entries; i++)
      {
        if (pat->palette[i].used_flag == 0)
          {
            out_data[0]=pat->palette[i].red;
            out_data[1]=pat->palette[i].green;
            out_data[2]=pat->palette[i].blue;
            out_data[3]=0x00;
            out_data[4]=(char)pat->palette[i].DMC_num;
            out_data[5]=(char)(pat->palette[i].DMC_num>>8);
            out_data[6]=0x00;
            out_data[7]=0x00;
            out_data[8]=0x00;
            out_data[9]=0x00;
            sprintf((char *)&out_data[10], "%s", pat->palette[i].name);
            out_data[35]=0x00;
            out_data[36]=(char)pat->palette[i].symbol;
            out_data[37]=0x00;
          }
        else
          {
            for (j=0; j<38; j++)
              out_data[j] = 0x00;
            out_data[3] = 0xFF;
          }
        write(pat_fd, out_data, 37);
      }

// This needs refining: it only does 1 stitch at a time (should do the 
// RLE compression used by the format)
    for (y=0; y < pat->header.height; y++)
      {
        out_data[0]=0; out_data[1]=-1;
        for (x=0; x < pat->header.width; x++)
          {
            testing = pat->stitches[y*pat->header.width + x].color;
            if (testing==-1) testing=0xFF;
            if (out_data[0]==0)
              {
                out_data[0]=1;
                out_data[1]=testing;
              }
            else
              {
                if (testing==out_data[1])
                  {
                    out_data[0]++;
                  }
                else
                  {
                    write(pat_fd, out_data, 2);
                    out_data[0]=1; out_data[1]=testing;
                  }
              }
          }
        if (out_data[0]!=0) write (pat_fd, out_data, 2);
      }

// And this does too, for the same reason
    for (y=0; y < pat->header.height; y++)
      {
        out_data[0]=0; out_data[1]=-1;
        for (x=0; x < pat->header.width; x++)
          {
            testing = pat->stitches[y*pat->header.width + x].type;
            if (out_data[0]==0)
              {
                out_data[0]=1;
                out_data[1]=testing;
              }
            else
              {
                if (testing==out_data[1])
                  {
                    out_data[0]++;
                  }
                else
                  {
                    write(pat_fd, out_data, 2);
                    out_data[0]=1; out_data[1]=testing;
                  }
              }
          }
        if (out_data[0]!=0) write (pat_fd, out_data, 2);
      }

// Okay, we're through this, so it's time for backstitches (ignore color)
    if (color_bstitches)
      {
        for (i=0; i < pat->header.num_bstitches; i++)
          {
            out_data[0]=(char)pat->bstitches[i].x1;
            out_data[1]=(char)(pat->bstitches[i].x1 >> 8);
            out_data[2]=(char)pat->bstitches[i].y1;
            out_data[3]=(char)(pat->bstitches[i].y1 >> 8);
            out_data[4]=(char)pat->bstitches[i].x2;
            out_data[5]=(char)(pat->bstitches[i].x2 >> 8);
            out_data[6]=(char)pat->bstitches[i].y2;
            out_data[7]=(char)(pat->bstitches[i].y2 >> 8);
            out_data[8]=(char)pat->bstitches[i].color;
            out_data[9]=(char)(pat->bstitches[i].color >> 8);
//          out_data[9]=0;                   // only 64 colors
            write(pat_fd, out_data, 10);
          }
      }
    else
      {
        for (i=0; i < pat->header.num_bstitches; i++)
          {
            out_data[0]=(char)pat->bstitches[i].x1;
            out_data[1]=(char)(pat->bstitches[i].x1>>8);
            out_data[2]=(char)pat->bstitches[i].y1;
            out_data[3]=(char)(pat->bstitches[i].y1>>8);
            out_data[4]=(char)pat->bstitches[i].x2;
            out_data[5]=(char)(pat->bstitches[i].x2>>8);
            out_data[6]=(char)pat->bstitches[i].y2;
            out_data[7]=(char)(pat->bstitches[i].y2>>8);
            write(pat_fd, out_data, 8);
          }
      }

    num_specials=0;
    for (i=0; i < pat->header.width*pat->header.height; i++)
      {
        if ((pat->stitches[i].type==7)||
            (pat->stitches[i].type==8))
          num_specials++;
      }

    if (num_specials > 0)
      {
        out_data[0]=(char)num_specials;
        out_data[1]=(char)(num_specials >> 8);
        write(pat_fd, out_data, 2);

        for (y=0; y < pat->header.height; y++)
          {
            for (x=0; x < pat->header.width; x++)
              {
                i = y*pat->header.width + x;
                if ((pat->stitches[i].type==7)||
                    (pat->stitches[i].type==8))
                  {
                    out_data[0] = (char)x;
                    out_data[1] = (char)(x >> 8);
                    out_data[2] = (char)y;
                    out_data[3] = (char)(y >> 8);
                    out_data[4] = (char)pat->stitches[y*pat->header.width 
                                                    + x].second;
                    out_data[5] = 0;
                    write(pat_fd, out_data, 6);
                  }
              }
          }
      }
    else
      {
        out_data[0]=0;
        out_data[1]=0;
        out_data[2]=0;
        write(pat_fd, out_data, 3);
      }

    close(pat_fd);

    return VPATIO_OK;

}

#endif

